/***************************************************************************
 *                           univariate.cpp                                *
 *  code to run univariate statistical maps                                *
 ***************************************************************************/

#include <blitz/array.h>
#include <time.h>
#include <random/uniform.h>
#include <fstream>

#include <omp.h>

#include <nifti1.h>
#include <nifti1_io.h>
#include <znzlib.h>

#include "read_model_matrices.h"
#include "read_images_nifti.h"
#include "write_images_nifti.h"
#include "threshold_mask.h"
#include "solve_stat.h"

using namespace std;
using namespace ranlib;


BZ_USING_NAMESPACE(blitz)

#include <string.h>
#include <math.h>

  int uni_analysis(ofstream *toFile, ifstream *fromFile) {

  int  num_scans, num_effects, num_confounds,fin_OK,num_contrasts, num_perm,norm,use_mask;
  int x_dim, y_dim, z_dim,zz_dim,  sep_files;
  int data_type, bits_pix;
  int a;
  float x_res, y_res, z_res, zz_res;
  float scale, intercept;
  char basename1[100], basename2[100], statname[100], maskname[100];

  *(fromFile) >> num_scans;
  *(fromFile) >> num_effects;
  *(fromFile) >> num_confounds;
  *(fromFile) >> norm;
  *(fromFile) >> num_contrasts;
  *(fromFile) >> num_perm;
  *(fromFile) >> sep_files;
  *(fromFile) >> use_mask;
  *(fromFile) >> basename1;
  *(fromFile) >> basename2;
  *(fromFile) >> statname;
  *(fromFile) >> maskname;
  *(toFile) << "number of scans: " << num_scans << endl;
  *(toFile) << "number of effects: " << num_effects << endl;
  *(toFile) << "number of confounds: " << num_confounds << endl;
  *(toFile) << "number of contrasts are: " << num_contrasts << endl;

  // allocate design matrices, list of scans

  Array<float,2> X1(num_scans, num_effects), X2(num_scans, num_confounds);
  Array<float,2> X(num_scans, num_effects+num_confounds);
  Array<float,2> C1;
  Array<float,1> normVal(num_scans);

  normVal=1;


  // read in design matrices
  char fname[500][200];
  if((fin_OK=read_design_mat(toFile,fromFile,X1,X2,X,norm,normVal,fname/*,num_scans,num_effects,num_confounds*/))!=0) {
    *(toFile) << "error reading design matrix" << endl;
    cout << "error reading design matrix, exiting" << endl;
    exit(1);
  }
  else *(toFile) << "design matrix read OK" << endl;

  // read in contrasts, if any
  if(num_contrasts>0) { 
    C1.resize(num_contrasts,num_effects+num_confounds);
    if( (fin_OK=read_contrast_mat(toFile, fromFile, C1))!=0) {
      *(toFile) << "error reading contrasts matrix" << endl;
      cout << "error reading contrasts matrix, exiting" << endl;
      exit(1);
    }
    else *(toFile) << "contrast matrix read OK" << endl;
  }

  // read in header so data array can be created
  nifti_image nim;
  nim.dim[1]=0;
  if( (fin_OK=read_image_hdr(toFile,fname[0],&nim))!=0) {
    *(toFile) << "error reading image header" << endl;
    cout << "error reading image header, exiting" << endl;
    exit(1);
  }
  else { 
      *(toFile) << "initial image header read OK" << endl;
      x_dim=nim.dim[1];
      y_dim=nim.dim[2];
      z_dim=nim.dim[3];
      zz_dim=nim.dim[4];
      x_res=nim.pixdim[1];
      y_res=nim.pixdim[2];
      z_res=nim.pixdim[3];
      zz_res=nim.pixdim[4];
      data_type=nim.datatype;
      bits_pix=nim.nbyper;
      scale=nim.scl_slope;
      intercept=nim.scl_inter;
  }

  // allocate array for all dependent variable images
  // even though univariate analysis, make array 5 dimensional to simplify read_images function later
  Array<float,5> images(x_dim,y_dim,z_dim,zz_dim,num_scans);
 
  //  read in files
  if((fin_OK=read_images(toFile,fname,images))!=0) {
    *(toFile) << "ERROR: could not read images " << endl;
    cout << "ERROR: could not read images, exiting" << endl;
    exit(1);
  }
  else *(toFile) << "dependent variable images read OK" << endl;

  // normalize if necessary
  if(norm==1) {
#pragma omp parallel for
    for(a=0; a<num_scans; a++) {
      images(Range::all(),Range::all(),Range::all(),Range::all(),a)/=normVal(a);
    }
  }
  Array<float,4> imageMask(x_dim,y_dim,z_dim,zz_dim);
  imageMask=1.;

  if(use_mask==1) {
    cout << "before reading imageMask" << endl;
    // try to read in mask file, average file, or reference brain
    if(( fin_OK=read_image(toFile,maskname,imageMask))!=0) {
        *(toFile) << "ERROR: could not read mask image" << endl;
        cout << "ERROR: could not read mask image, exiting" << endl;
        exit(1);
    }
    if( (fin_OK=threshold_mask(imageMask,maskname))!=0) {
          *(toFile) << "ERROR: could not threshold mask image" << endl;
          cout << "ERROR: could not threshold mask image, exiting" << endl;
        exit(1);
    }
    else {
        *(toFile) << "mask image thresholded successfully" << endl;
    }
  }
  // Here add code to mask the ind_images
  clock_t start, end;
  if(use_mask==1) {
  for(a=0; a<num_scans; a++) {
    images(Range::all(),Range::all(),Range::all(),Range::all(),a)*=imageMask;
  }
  }

  // solve problem the first time
  //  if ( (fin_OK=solve_stat(toFile,images,X,X2,C1,

  // allocate arrays for solutions and statistics
  Array<float,5> Bimage(x_dim, y_dim, z_dim, zz_dim, num_effects+num_confounds);
  Array<float,5> B2image(x_dim, y_dim, z_dim, zz_dim, num_confounds);
  Array<float,3> Stat_image(x_dim,y_dim,z_dim);
  Array<float,5> Stat_param_image(x_dim,y_dim,z_dim,zz_dim,num_effects+num_confounds);
  Array<float,5> Contrast_image(x_dim,y_dim,z_dim,zz_dim,num_contrasts);
  Array<float,5> F_contrast_image(x_dim,y_dim,z_dim,zz_dim,num_contrasts);
  Array<float,4> Stat_R2_image(x_dim,y_dim,z_dim,zz_dim);
  Uniform<float> unif;
  unif.seed((unsigned int)21466 /*time(0)*/);

  // now compute solution and statistics
  float df1, df2, tdf, cdf1, cdf2;
  df1=df2=tdf=cdf1=cdf2=0;
  cout << "getting read to call solve_stat" << endl;
  if( (fin_OK=solve_stat(toFile,images,X,X2,C1,Bimage,B2image,Stat_image,&df1,&df2,Stat_param_image,&tdf,Contrast_image,F_contrast_image,&cdf1,&cdf2,Stat_R2_image,unif,0))!=0) {
    *(toFile) << "ERROR: could not solve model" << endl;
    cout << "ERROR: could not solve model, exiting" << endl;
    exit(1);
  }

  // write Bimage and B2image to files
  if((fin_OK=writeBetaImage(toFile, Bimage,&nim /*x_res,y_res,z_res,zz_res*/,fname[0],basename1))!=0) {
    *(toFile) << "ERROR: could not write full solution to image files" << endl;
    cout << "ERROR: could not write full solution to image files, exiting" << endl;
    exit(1);
  }
  if((fin_OK=writeBetaImage(toFile, B2image,&nim /*x_res,y_res,z_res,zz_res*/,fname[0],basename2))!=0) {
    *(toFile) << "ERROR: could not write reduced solution to image files" << endl;
    cout << "ERROR: could not write reduced solution to image files, exiting" << endl;
    exit(1);
  }
  
  // write Stat_image, Stat_param_image, R2image to files
  if((fin_OK=writeStatImage(toFile, Stat_image,&nim /*x_res,y_res,z_res,zz_res*/,fname[0],statname,&df1,&df2))!=0) {
    *(toFile) << "ERROR: could not write F statistic file" << endl;
    cout << "ERROR: could not write F statistic file, exiting" << endl;
    exit(1);
  }
  if((fin_OK=writeStatParamImage(toFile, Stat_param_image,&nim /*x_res,y_res,z_res,zz_res*/,fname[0],statname,&tdf))!=0) {
    *(toFile) << "ERROR: could not write F statistic parameter file" << endl;
    cout << "ERROR: could not write F statistic parameter file, exiting" << endl;
    exit(1);
  }
  if((fin_OK=writeR2Image(toFile, Stat_R2_image,&nim /*x_res,y_res,z_res,zz_res*/,fname[0],statname))!=0) {
    *(toFile) << "ERROR: could not write R2 file" << endl;
    cout << "ERROR: could not write R2 file, exiting" << endl;
    exit(1);
  }
  if(C1.extent(firstDim)!=0) {
    if((fin_OK=writeContrastImage(toFile, Contrast_image,&nim /*x_res,y_res,z_res,zz_res*/,fname[0],basename1))!=0) {
      *(toFile) << "ERROR: could not write contrast image" << endl;
    }
    if((fin_OK=writeContrastFImage(toFile, F_contrast_image,&nim /*x_res,y_res,z_res,zz_res*/,fname[0],statname,&cdf1,&cdf2))!=0) {
      *(toFile) << "ERROR: could not write contrast F-statistic image" << endl;
    }
  }
  // at some point, maybe compute roughness of image??
  // need to look at Worsley code, since my old versions didn't seem to work
  
  // compute permutatations
  if( num_perm > 0) {
    Array<float,1> Stat_Fmax(num_perm);
    Array<float,1> Stat_Fmin(num_perm);
    Array<float,3> Stat_param_Tmax(zz_dim,num_effects+num_confounds, num_perm);
    Array<float,3> Stat_param_Tmin(zz_dim,num_effects+num_confounds, num_perm);
    Array<float,3> Stat_contrast_Fmax(zz_dim,num_contrasts, num_perm);
    Array<float,3> Stat_contrast_Fmin(zz_dim,num_contrasts, num_perm);
    *(toFile) << "allocated permutation matrices OK" << endl;



    // after Memorial Day, adjust code to permute X matrix every time and save max statistics, etc.
    start=clock();
    fin_OK=permutations(toFile,images,X,X2,C1,Bimage,B2image,Stat_image,&df1,&df2,Stat_param_image,&tdf,Contrast_image,F_contrast_image,&cdf1,&cdf2,Stat_R2_image,unif,num_perm,Stat_Fmax,Stat_Fmin,Stat_param_Tmax,Stat_param_Tmin,Stat_contrast_Fmax,Stat_contrast_Fmin);
    end=clock();
    cout << "permutation time " << end-start << endl;
  }
  
  return(0);
}

int uni_analysis_spvar(ofstream *toFile, ifstream *fromFile,int num_spvar_indep_var) {

  int  num_scans, num_effects, num_confounds,fin_OK,num_contrasts, num_perm,norm, use_mask;
  int x_dim, y_dim, z_dim,zz_dim,  sep_files;
  int data_type, bits_pix;
  int a,b;
  float x_res, y_res, z_res, zz_res;
  float scale, intercept;
  char basename1[100], basename2[100], statname[100], maskname[100];
 
  //  cout << "num spatially varying variables " << num_spvar_indep_var << endl;

  *(fromFile) >> num_scans;
  *(fromFile) >> num_effects;
  *(fromFile) >> num_confounds;
  *(fromFile) >> norm;
  *(fromFile) >> num_contrasts;
  *(fromFile) >> num_perm;
  *(fromFile) >> sep_files;
  *(fromFile) >> use_mask;
  *(fromFile) >> basename1;
  *(fromFile) >> basename2;
  *(fromFile) >> statname;
  *(fromFile) >> maskname;
  *(toFile) << "number of scans: " << num_scans << endl;
  *(toFile) << "number of effects: " << num_effects << endl;
  *(toFile) << "number of confounds: " << num_confounds << endl;
  *(toFile) << "number of contrasts are: " << num_contrasts << endl;

  // allocate design matrices, list of scans

  Array<float,2> X1(num_scans, num_effects), X2(num_scans, num_confounds);
  Array<float,2> X(num_scans, num_effects+num_confounds);
  Array<float,2> C1;
  Array<float,1> normVal(num_scans);

  normVal=1;


  // read in design matrices
  char fname[500][200],fnamesp[500][10][200];
  if((fin_OK=read_design_mat_sp(toFile,fromFile,X1,X2,X,norm,normVal,fname,fnamesp,num_spvar_indep_var))!=0) {
    *(toFile) << "error reading design matrix" << endl;
    cout << "error reading design matrix, exiting" << endl;
    exit(1);
  }
  else *(toFile) << "design matrix read OK" << endl;

  // read in contrasts, if any
  if(num_contrasts>0) { 
    C1.resize(num_contrasts,num_effects+num_confounds);
    if( (fin_OK=read_contrast_mat(toFile, fromFile, C1))!=0) {
      *(toFile) << "error reading contrasts matrix" << endl;
      cout << "error reading contrasts matrix, exiting" << endl;
      exit(1);
    }
    else *(toFile) << "contrast matrix read OK" << endl;
  }

  // read in header so data array can be created
  nifti_image nim;
  nim.dim[1]=0;
  if( (fin_OK=read_image_hdr(toFile,fname[0],&nim))!=0) {
    *(toFile) << "error reading image header" << endl;
    cout << "error reading image header, exiting" << endl;
    exit(1);
  }
  else {
      *(toFile) << "initial image header read OK" << endl;
      x_dim=nim.dim[1];
      y_dim=nim.dim[2];
      z_dim=nim.dim[3];
      zz_dim=nim.dim[4];
      x_res=nim.pixdim[1];
      y_res=nim.pixdim[2];
      z_res=nim.pixdim[3];
      zz_res=nim.pixdim[4];
      data_type=nim.datatype;
      bits_pix=nim.nbyper;
      scale=nim.scl_slope;
      intercept=nim.scl_inter;
  }

  // allocate array for all dependent variable images
  // even though univariate analysis, make array 5 dimensional to simplify read_images function later
  Array<float,5> images(x_dim,y_dim,z_dim,zz_dim,num_scans);
  Array<float,6> ind_images(x_dim,y_dim,z_dim,zz_dim,num_spvar_indep_var,num_scans);
 
  //  read in files
  if((fin_OK=read_images(toFile,fname,images))!=0) {
    *(toFile) << "ERROR: could not read images " << endl;
    cout << "ERROR: could not read images, exiting" << endl;
    exit(1);
  }
  else *(toFile) << "dependent variable images read OK" << endl;

  // read in spatially varying independent images
  if((fin_OK=read_ind_images(toFile,fnamesp,ind_images))!=0) {
    *(toFile) << "ERROR: could not read independent variable images " << endl;
    cout << "ERROR: could not read independent variable images, exiting" << endl;
    exit(1);
  }

  // normalize if necessary
  if(norm==1) {
    for(a=0; a<num_scans; a++) {
      images(Range::all(),Range::all(),Range::all(),Range::all(),a)/=normVal(a);
    }
  }
  Array<float,4> imageMask(x_dim,y_dim,z_dim,zz_dim);
  imageMask=1.;

  if(use_mask==1) {
    // try to read in mask file, average file, or reference brain
    if(( fin_OK=read_image(toFile,maskname,imageMask))!=0) {
        *(toFile) << "ERROR: could not read mask image" << endl;
    }
    if( (fin_OK=threshold_mask(imageMask,maskname))!=0) {
        *(toFile) << "ERROR: could not threshold mask image" << endl;
        cout << "ERROR: could not threshole mask image, exiting" << endl;
        exit(1);
    }
    else *(toFile) << "mask image thresholded successfully" << endl;
  }
  // Here add code to mask the ind_images and images

    if(use_mask==1) {
  for(a=0; a<num_scans; a++) {
    images(Range::all(),Range::all(),Range::all(),Range::all(),a)*=imageMask;
  }
  for(a=0; a<num_scans; a++) {
    for (b=0; b<num_spvar_indep_var; b++) {
      ind_images(Range::all(),Range::all(),Range::all(),Range::all(),b,a)*=imageMask;
    }
  }
    }

  // allocate arrays for solutions and statistics
  Array<float,5> Bimage(x_dim, y_dim, z_dim, zz_dim, num_effects+num_confounds);
  Array<float,5> B2image(x_dim, y_dim, z_dim, zz_dim, num_confounds);
  Array<float,3> Stat_image(x_dim,y_dim,z_dim);
  Array<float,5> Stat_param_image(x_dim,y_dim,z_dim,zz_dim,num_effects+num_confounds);
  Array<float,5> Contrast_image(x_dim,y_dim,z_dim,zz_dim,num_contrasts);
  Array<float,5> F_contrast_image(x_dim,y_dim,z_dim,zz_dim,num_contrasts);
  Array<float,4> Stat_R2_image(x_dim,y_dim,z_dim,zz_dim);
  Uniform<float> unif;
  unif.seed((unsigned int)21466 /*time(0)*/);

  // now compute solution and statistics
  float df1, df2, tdf, cdf1, cdf2;
  df1=df2=tdf=cdf1=cdf2=0;
  if( (fin_OK=solve_stat_spvar(toFile,images,ind_images,num_spvar_indep_var,X,X2,C1,Bimage,B2image,Stat_image,&df1,&df2,Stat_param_image,&tdf,Contrast_image,F_contrast_image,&cdf1,&cdf2,Stat_R2_image,unif,0))!=0) {
    *(toFile) << "ERROR: could not solve model" << endl;
    cout << "ERROR: could not solve model, exiting" << endl;
    exit(1);
  }
  
  // write Bimage and B2image to files
  if((fin_OK=writeBetaImage(toFile, Bimage,&nim /*x_res,y_res,z_res,zz_res*/,fname[0],basename1))!=0) {
    *(toFile) << "ERROR: could not write full solution to image files" << endl;
    cout << "ERROR: could not write full solution to image files, exiting" << endl;
    exit(1);
  }
  if((fin_OK=writeBetaImage(toFile, B2image,&nim /*x_res,y_res,z_res,zz_res*/,fname[0],basename2))!=0) {
    *(toFile) << "ERROR: could not write reduced solution to image files" << endl;
    cout << "ERROR: could not write reduced solution to image files, exiting" << endl;
    exit(1);
  }
  
  // write Stat_image, Stat_param_image, R2image to files
  if((fin_OK=writeStatImage(toFile, Stat_image,&nim /*x_res,y_res,z_res,zz_res*/,fname[0],statname,&df1,&df2))!=0) {
    *(toFile) << "ERROR: could not write F statistic file" << endl;
    cout << "ERROR: could not write F statistic file, exiting " << endl;
    exit(1);
  }
  if((fin_OK=writeStatParamImage(toFile, Stat_param_image,&nim /*x_res,y_res,z_res,zz_res*/,fname[0],statname,&tdf))!=0) {
    *(toFile) << "ERROR: could not write F statistic parameter file" << endl;
    cout << "ERROR: could not write F statistic parameter file, exiting" << endl;
    exit(1);
  }
  if((fin_OK=writeR2Image(toFile, Stat_R2_image,&nim /*x_res,y_res,z_res,zz_res*/,fname[0],statname))!=0) {
    *(toFile) << "ERROR: could not write R2s file" << endl;
    cout << "ERROR: could not write R2 file, exiting" << endl;
    exit(1);
  }
  if(C1.extent(firstDim)!=0) {
    if((fin_OK=writeContrastImage(toFile, Contrast_image,&nim /*x_res,y_res,z_res,zz_res*/,fname[0],basename1))!=0) {
      *(toFile) << "ERROR: could not write contrast image" << endl;
    }
    if((fin_OK=writeContrastFImage(toFile, F_contrast_image,&nim /*x_res,y_res,z_res,zz_res*/,fname[0],statname,&cdf1,&cdf2))!=0) {
      *(toFile) << "ERROR: could not write contrast F-statistic image" << endl;
    }
  }
  // at some point, maybe compute roughness of image??
  // need to look at Worsley code, since my old versions didn't seem to work
  
  // compute permutatations
  if(num_perm>0) {
    Array<float,1> Stat_Fmax(num_perm);
    Array<float,1> Stat_Fmin(num_perm);
    Array<float,3> Stat_param_Tmax(zz_dim,num_effects+num_confounds, num_perm);
    Array<float,3> Stat_param_Tmin(zz_dim,num_effects+num_confounds, num_perm);
    Array<float,3> Stat_contrast_Fmax(zz_dim,num_contrasts, num_perm);
    Array<float,3> Stat_contrast_Fmin(zz_dim,num_contrasts, num_perm);
    *(toFile) << "allocated permutation matrices OK" << endl;



    // after Memorial Day, adjust code to permute X matrix every time and save max statistics, etc.

    fin_OK=permutations_spvar(toFile,images,ind_images,num_spvar_indep_var,X,X2,C1,Bimage,B2image,Stat_image,&df1,&df2,Stat_param_image,&tdf,Contrast_image,F_contrast_image,&cdf1,&cdf2,Stat_R2_image,unif,num_perm,Stat_Fmax,Stat_Fmin,Stat_param_Tmax,Stat_param_Tmin,Stat_contrast_Fmax,Stat_contrast_Fmin);
  }
  
  return(0);
}

int uni_analysis_staticdv_spvar(ofstream *toFile, ifstream *fromFile,int num_spvar_indep_var) {
  int  num_obs, num_effects, num_confounds,fin_OK,num_contrasts, num_perm,norm, sep_files, use_mask;
  int x_dim, y_dim, z_dim,zz_dim;
  int data_type, bits_pix;
  int a,b;
  float x_res, y_res, z_res, zz_res;
  float scale, intercept;
  char basename1[100], basename2[100], statname[100], maskname[100];
 
  //  cout << "num spatially varying variables " << num_spvar_indep_var << endl;

  *(fromFile) >> num_obs;
  *(fromFile) >> num_effects;
  *(fromFile) >> num_confounds;
  *(fromFile) >> norm;
  *(fromFile) >> num_contrasts;
  *(fromFile) >> num_perm;
  *(fromFile) >> sep_files;
  *(fromFile) >> use_mask;
  *(fromFile) >> basename1;
  *(fromFile) >> basename2;
  *(fromFile) >> statname;
  *(fromFile) >> maskname;
  *(toFile) << "number of observations: " << num_obs << endl;
  *(toFile) << "number of effects: " << num_effects << endl;
  *(toFile) << "number of confounds: " << num_confounds << endl;
  *(toFile) << "number of contrasts are: " << num_contrasts << endl;
  
  // allocate design matrices, list of scans

  Array<float,2> X1(num_obs, num_effects), X2(num_obs, num_confounds);
  Array<float,2> X(num_obs, num_effects+num_confounds);
  Array<float,2> C1;
  Array<float,1> normVal(num_obs);
  Array<float,2> Y(1,num_obs);

  normVal=1;

  
  // read in design matrices
  char fnamesp[500][10][200]; 
  (*toFile) << " about to read" << endl;
  
  if((fin_OK=read_design_mat_static_sp(toFile,fromFile,X1,X2,X,norm,normVal,Y,fnamesp,num_spvar_indep_var))!=0) {
    *(toFile) << "error reading design matrix" << endl;
    cout << "error reading design matrix, exiting" << endl;
    exit(1);
  }
  else *(toFile) << "design matrix read OK" << endl;
  
  // read in contrasts, if any
  if(num_contrasts>0) { 
    C1.resize(num_contrasts,num_effects+num_confounds);
    if( (fin_OK=read_contrast_mat(toFile, fromFile, C1))!=0) {
      *(toFile) << "error reading contrasts matrix" << endl;
      cout << "error reading contrasts matrix, exiting" << endl;
      exit(1);
    }
    else *(toFile) << "contrast matrix read OK" << endl;
  }
  
  // read in header so data array can be created
  nifti_image nim;
  nim.dim[1]=0;
  cout << "about to read in first file?" << endl;
  cout << fnamesp[0][0] << endl;
  if( (fin_OK=read_image_hdr(toFile,fnamesp[0][0],&nim))!=0) {
    *(toFile) << "error reading image header" << endl;
    cout << "error reading image header, exiting" << endl;
    exit(1);
  }
  else {
      *(toFile) << "initial image header read OK" << endl;
      x_dim=nim.dim[1];
      y_dim=nim.dim[2];
      z_dim=nim.dim[3];
      zz_dim=nim.dim[4];
      x_res=nim.pixdim[1];
      y_res=nim.pixdim[2];
      z_res=nim.pixdim[3];
      zz_res=nim.pixdim[4];
      data_type=nim.datatype;
      bits_pix=nim.nbyper;
      scale=nim.scl_slope;
      intercept=nim.scl_inter;
  }

  // even though univariate analysis, make array 6 dimensional to simplify read_images function later
  Array<float,6> ind_images(x_dim,y_dim,z_dim,zz_dim,num_spvar_indep_var,num_obs);
  
  // read in spatially varying independent images
  if((fin_OK=read_ind_images(toFile,fnamesp,ind_images))!=0) {
    *(toFile) << "ERROR: could not read independent variable images " << endl;
    cout << "ERROR: could not read independent variable images, exiting" << endl;
    exit(1);
  }
  Array<float,4> imageMask(x_dim,y_dim,z_dim,zz_dim);
  imageMask=1.;

  if(use_mask==1) {
    // try to read in mask file, average file, or reference brain
      if(( fin_OK=read_image(toFile,maskname,imageMask))!=0) {
        *(toFile) << "ERROR: could not read mask image" << endl;
        }
    if( (fin_OK=threshold_mask(imageMask,maskname))!=0) {
        *(toFile) << "ERROR: could not threshold mask image" << endl;
        cout << "ERROR: could not threshold mask image, exiting" << endl;
        exit(1);
    }
    else *(toFile) << "mask image thresholded successfully" << endl;
  }
  // Here add code to mask the ind_images
    if(use_mask==1) {
  for(a=0; a<num_obs; a++) {
    for (b=0; b<num_spvar_indep_var; b++) {
      ind_images(Range::all(),Range::all(),Range::all(),Range::all(),b,a)*=imageMask;
    }
  }
    }
  
  
  // allocate arrays for solutions and statistics
  Array<float,5> Bimage(x_dim, y_dim, z_dim, zz_dim, num_effects+num_confounds);
  Array<float,5> B2image(x_dim, y_dim, z_dim, zz_dim, num_confounds);
  Array<float,3> Stat_image(x_dim,y_dim,z_dim);
  Array<float,5> Stat_param_image(x_dim,y_dim,z_dim,zz_dim,num_effects+num_confounds);
  Array<float,5> Contrast_image(x_dim,y_dim,z_dim,zz_dim,num_contrasts);
  Array<float,5> F_contrast_image(x_dim,y_dim,z_dim,zz_dim,num_contrasts);
  Array<float,4> Stat_R2_image(x_dim,y_dim,z_dim,zz_dim);
  Uniform<float> unif;
  unif.seed((unsigned int)time(0));
  
  // now compute solution and statistics
  float df1, df2, tdf, cdf1, cdf2;
  df1=df2=tdf=cdf1=cdf2=0;

  if( (fin_OK=solve_stat_static_spvar(toFile,imageMask,Y,ind_images,num_spvar_indep_var,X,X2,C1,Bimage,B2image,Stat_image,&df1,&df2,Stat_param_image,&tdf,Contrast_image,F_contrast_image,&cdf1,&cdf2,Stat_R2_image,unif,0))!=0) {
    *(toFile) << "ERROR: could not solve model" << endl;
    cout << "ERROR: could not solve model, exiting" << endl;
    exit(1);
  }

  // now write results to files, should be the same as for other spvar
  // write Bimage and B2image to files
  if((fin_OK=writeBetaImage(toFile, Bimage,&nim /*x_res,y_res,z_res,zz_res*/,maskname,basename1))!=0) {
    *(toFile) << "ERROR: could not write full solution to image files" << endl;
    cout << "ERROR: could not write full solution to image files, exiting" << endl;
    exit(1);
  }
  if((fin_OK=writeBetaImage(toFile, B2image,&nim /*x_res,y_res,z_res,zz_res*/,maskname,basename2))!=0) {
    *(toFile) << "ERROR: could not write reduced solution to image files" << endl;
    cout << "ERROR: could not write reduced solution to image files, exiting" << endl;
    exit(1);
  }

  // write Stat_image, Stat_param_image, R2image to files
  if((fin_OK=writeStatImage(toFile, Stat_image,&nim /*x_res,y_res,z_res,zz_res*/,maskname,statname,&df1,&df2))!=0) {
    *(toFile) << "ERROR: could not write F statistic file" << endl;
    cout << "ERROR: could not write F statistic file, exiting" << endl;
    exit(1);
  }
  if((fin_OK=writeStatParamImage(toFile, Stat_param_image,&nim /*x_res,y_res,z_res,zz_res*/,maskname,statname,&tdf))!=0) {
    *(toFile) << "ERROR: could not write F statistic parameter file" << endl;
    cout << "ERROR: could not write F statistic parameter file, exiting" << endl;
    exit(1);
  }
  if((fin_OK=writeR2Image(toFile, Stat_R2_image,&nim /*x_res,y_res,z_res,zz_res*/,maskname,statname))!=0) {
    *(toFile) << "ERROR: could not write R2 file" << endl;
    cout << "ERROR: could not write R2 file, exiting" << endl;
    exit(1);
  }

  if(C1.extent(firstDim)!=0) {
    if((fin_OK=writeContrastImage(toFile, Contrast_image,&nim /*x_res,y_res,z_res,zz_res*/,maskname,basename1))!=0) {
      *(toFile) << "ERROR: could not write contrast image" << endl;
    }
    if((fin_OK=writeContrastFImage(toFile, F_contrast_image,&nim /*x_res,y_res,z_res,zz_res*/,maskname,statname,&cdf1,&cdf2))!=0) {
      *(toFile) << "ERROR: could not write contrast F-statistic image" << endl;
    }
  }

  // setup permutation matrices, run permutations
  // compute permutatations
  if(num_perm>0) {
    Array<float,1> Stat_Fmax(num_perm);
    Array<float,1> Stat_Fmin(num_perm);
    Array<float,3> Stat_param_Tmax(zz_dim,num_effects+num_confounds, num_perm);
    Array<float,3> Stat_param_Tmin(zz_dim,num_effects+num_confounds, num_perm);
    Array<float,3> Stat_contrast_Fmax(zz_dim,num_contrasts, num_perm);
    Array<float,3> Stat_contrast_Fmin(zz_dim,num_contrasts, num_perm);
    *(toFile) << "allocated permutation matrices OK" << endl;

    fin_OK=permutations_spvar_static(toFile,imageMask,Y,ind_images,num_spvar_indep_var,X,X2,C1,Bimage,B2image,Stat_image,&df1,&df2,Stat_param_image,&tdf,Contrast_image,F_contrast_image,&cdf1,&cdf2,Stat_R2_image,unif,num_perm,Stat_Fmax,Stat_Fmin,Stat_param_Tmax,Stat_param_Tmin,Stat_contrast_Fmax,Stat_contrast_Fmin);
  }
  cout << "finished univariate analysis" << endl;
  *(toFile) << "finished univariate analysis" << endl;

  (*toFile) << "ending program" << endl;
  return(0);
}
